"""Tests for serialization helpers."""

import json
import uuid
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict

import pytest

from agentops.helpers.serialization import (
    filter_unjsonable,
    is_jsonable,
    model_to_dict,
    safe_serialize,
    serialize_uuid,
)


# Define test models and data structures
class SampleEnum(Enum):
    ONE = 1
    TWO = 2
    THREE = "three"


class SimpleModel:
    """A simple class with __dict__ but no model_dump or dict method."""

    def __init__(self, value: str):
        self.value = value


class ModelWithToJson:
    """A class that implements to_json method."""

    def __init__(self, data: Dict):
        self.data = data

    def to_json(self):
        return self.data


class PydanticV1Model:
    """Mock Pydantic v1 model with dict method."""

    def __init__(self, **data):
        self.__dict__.update(data)

    def dict(self):
        return self.__dict__


class PydanticV2Model:
    """Mock Pydantic v2 model with model_dump method."""

    def __init__(self, **data):
        self.__dict__.update(data)

    def model_dump(self):
        return self.__dict__


class ModelWithParse:
    """Mock model with parse method."""

    def __init__(self, data):
        self.data = data

    def parse(self):
        return self.data


class ModelWithoutDict:
    """A class without __dict__ attribute."""

    __slots__ = ["value"]

    def __init__(self, value: str):
        self.value = value


# Define test cases for is_jsonable
class TestIsJsonable:
    def test_jsonable_types(self):
        """Test that jsonable types return True."""
        jsonable_objects = [
            "string",
            "",
            123,
            123.45,
            True,
            False,
            None,
            [1, 2, 3],
            {"key": "value"},
            [],
            {},
        ]

        for obj in jsonable_objects:
            assert is_jsonable(obj) is True

    def test_unjsonable_types(self):
        """Test that unjsonable types return False."""
        unjsonable_objects = [
            datetime.now(),
            uuid.uuid4(),
            Decimal("123.45"),
            {1, 2, 3},  # set
            SampleEnum.ONE,
            lambda x: x,  # function
            object(),  # generic object
        ]

        for obj in unjsonable_objects:
            assert is_jsonable(obj) is False

    def test_circular_reference(self):
        """Test that circular references are not jsonable."""
        a = {}
        b = {}
        a["b"] = b
        b["a"] = a

        # The current implementation doesn't handle ValueError from circular references
        # So this will raise an exception instead of returning False
        with pytest.raises(ValueError, match="Circular reference detected"):
            is_jsonable(a)


# Define test cases for filter_unjsonable
class TestFilterUnjsonable:
    def test_filter_simple_dict(self):
        """Test filtering of simple dictionary."""
        input_dict = {
            "string": "value",
            "number": 42,
            "list": [1, 2, 3],
            "dict": {"nested": "value"},
            "uuid": uuid.uuid4(),
            "datetime": datetime.now(),
            "set": {1, 2, 3},
        }

        result = filter_unjsonable(input_dict)

        # Check that jsonable values are preserved
        assert result["string"] == "value"
        assert result["number"] == 42
        assert result["list"] == [1, 2, 3]
        assert result["dict"] == {"nested": "value"}

        # Check that unjsonable values are converted to strings or empty strings
        assert isinstance(result["uuid"], str)
        assert result["datetime"] == ""
        assert result["set"] == ""

    def test_filter_nested_dict(self):
        """Test filtering of nested dictionaries."""
        input_dict = {
            "level1": {
                "level2": {
                    "uuid": uuid.uuid4(),
                    "string": "preserved",
                    "datetime": datetime.now(),
                }
            },
            "list_with_unjsonable": [
                {"uuid": uuid.uuid4()},
                "string",
                datetime.now(),
            ],
        }

        result = filter_unjsonable(input_dict)

        # Check nested structure is preserved
        assert result["level1"]["level2"]["string"] == "preserved"
        assert isinstance(result["level1"]["level2"]["uuid"], str)
        assert result["level1"]["level2"]["datetime"] == ""

        # Check list filtering
        assert result["list_with_unjsonable"][1] == "string"
        assert isinstance(result["list_with_unjsonable"][0]["uuid"], str)
        assert result["list_with_unjsonable"][2] == ""

    def test_filter_list(self):
        """Test filtering of lists."""
        input_list = [
            "string",
            42,
            uuid.uuid4(),
            datetime.now(),
            [1, 2, uuid.uuid4()],
            {"uuid": uuid.uuid4()},
        ]

        result = filter_unjsonable(input_list)

        assert result[0] == "string"
        assert result[1] == 42
        assert isinstance(result[2], str)  # UUID converted to string
        assert result[3] == ""  # datetime converted to empty string
        assert isinstance(result[4][2], str)  # nested UUID converted to string
        assert isinstance(result[5]["uuid"], str)  # nested UUID converted to string

    def test_filter_empty_structures(self):
        """Test filtering of empty structures."""
        assert filter_unjsonable({}) == {}
        assert filter_unjsonable([]) == []
        assert filter_unjsonable({"empty": {}}) == {"empty": {}}


# Define test cases for serialize_uuid
class TestSerializeUuid:
    def test_serialize_uuid(self):
        """Test UUID serialization."""
        test_uuid = uuid.uuid4()
        result = serialize_uuid(test_uuid)

        assert isinstance(result, str)
        assert result == str(test_uuid)

    def test_serialize_uuid_string(self):
        """Test that UUID string representation is correct."""
        test_uuid = uuid.UUID("00000000-0000-0000-0000-000000000001")
        result = serialize_uuid(test_uuid)

        assert result == "00000000-0000-0000-0000-000000000001"


# Define test cases for safe_serialize
class TestSafeSerialize:
    def test_strings_returned_untouched(self):
        """Test that strings are returned untouched."""
        test_strings = [
            "simple string",
            "",
            "special chars: !@#$%^&*()",
            '{"json": "string"}',  # JSON as a string
            "[1, 2, 3]",  # JSON array as a string
            "line 1\nline 2",  # String with newlines
        ]

        for input_str in test_strings:
            # The string should be returned exactly as is
            assert safe_serialize(input_str) == input_str

    def test_complex_objects_serialized(self):
        """Test that complex objects are properly serialized."""
        test_cases = [
            # Test case, expected serialized form (or None for dict check)
            ({"key": "value"}, '{"key": "value"}'),
            ([1, 2, 3], "[1, 2, 3]"),
            (123, "123"),
            (123.45, "123.45"),
            (True, "true"),
            (False, "false"),
            (None, "null"),
        ]

        for input_obj, expected in test_cases:
            result = safe_serialize(input_obj)
            if expected is not None:
                # Check exact match for simple cases
                assert json.loads(result) == json.loads(expected)
            else:
                # For complex cases just verify it's valid JSON
                assert isinstance(result, str)
                assert json.loads(result) is not None

    def test_pydantic_models(self):
        """Test serialization of Pydantic-like models."""
        # V1 model with dict()
        v1_model = PydanticV1Model(name="test", value=42)
        v1_result = safe_serialize(v1_model)
        assert json.loads(v1_result) == {"name": "test", "value": 42}

        # V2 model with model_dump()
        v2_model = PydanticV2Model(name="test", value=42)
        v2_result = safe_serialize(v2_model)
        assert json.loads(v2_result) == {"name": "test", "value": 42}

        # Note: parse() method is currently not implemented due to recursion issues
        # See TODO in serialization.py

    def test_special_types(self):
        """Test serialization of special types using AgentOpsJSONEncoder."""
        test_cases = [
            # Datetime
            (datetime(2023, 1, 1, 12, 0, 0), '"2023-01-01T12:00:00"'),
            # UUID
            (uuid.UUID("00000000-0000-0000-0000-000000000001"), '"00000000-0000-0000-0000-000000000001"'),
            # Decimal
            (Decimal("123.45"), '"123.45"'),
            # Set
            ({1, 2, 3}, "[1, 2, 3]"),
            # Enum
            (SampleEnum.ONE, "1"),
            (SampleEnum.THREE, '"three"'),
            # Class with to_json
            (ModelWithToJson({"key": "value"}), '{"key": "value"}'),
        ]

        for input_obj, expected in test_cases:
            result = safe_serialize(input_obj)

            # Handle list comparison for sets where order might vary
            if isinstance(input_obj, set):
                assert sorted(json.loads(result)) == sorted(json.loads(expected))
            else:
                assert json.loads(result) == json.loads(expected)

    def test_nested_objects(self):
        """Test serialization of nested objects."""
        nested_obj = {
            "string": "value",
            "number": 42,
            "list": [1, 2, {"inner": "value"}],
            "dict": {"inner": {"deeper": [1, 2, 3]}},
            "model": PydanticV2Model(name="test"),
        }

        result = safe_serialize(nested_obj)

        # Verify it's valid JSON
        parsed = json.loads(result)
        assert parsed["string"] == "value"
        assert parsed["number"] == 42
        assert parsed["list"][2]["inner"] == "value"
        assert parsed["dict"]["inner"]["deeper"] == [1, 2, 3]

        # Just verify we have the model in some form
        assert "model" in parsed
        # And verify it contains the expected data in some form
        assert "test" in str(parsed["model"])

    def test_fallback_to_str(self):
        """Test fallback to str() for unserializable objects."""

        class Unserializable:
            def __str__(self):
                return "Unserializable object"

        obj = Unserializable()
        result = safe_serialize(obj)
        # The string is wrapped in quotes because it's serialized as a JSON string
        assert result == '"Unserializable object"'

    def test_serialization_error_handling(self):
        """Test handling of serialization errors."""

        # Create an object that causes JSON serialization to fail
        class BadObject:
            def __init__(self):
                self.recursive = None

            def __getitem__(self, key):
                # This will cause infinite recursion during JSON serialization
                return self.recursive

            def __str__(self):
                return "BadObject representation"

        bad_obj = BadObject()
        bad_obj.recursive = bad_obj

        result = safe_serialize(bad_obj)
        assert result == '"BadObject representation"'

    def test_value_error_handling(self):
        """Test handling of ValueError during JSON serialization."""

        # Create an object that causes a ValueError during JSON serialization
        class ValueErrorObject:
            def to_json(self):
                raise ValueError("Cannot serialize this object")

            def __str__(self):
                return "ValueErrorObject representation"

        obj = ValueErrorObject()
        result = safe_serialize(obj)
        assert result == "ValueErrorObject representation"


class TestModelToDict:
    def test_none_returns_empty_dict(self):
        """Test that None returns an empty dict."""
        assert model_to_dict(None) == {}

    def test_dict_returns_unchanged(self):
        """Test that a dict is returned unchanged."""
        test_dict = {"key": "value"}
        assert model_to_dict(test_dict) is test_dict

    def test_pydantic_models(self):
        """Test conversion of Pydantic-like models to dicts."""
        # V1 model with dict()
        v1_model = PydanticV1Model(name="test", value=42)
        assert model_to_dict(v1_model) == {"name": "test", "value": 42}

        # V2 model with model_dump()
        v2_model = PydanticV2Model(name="test", value=42)
        assert model_to_dict(v2_model) == {"name": "test", "value": 42}

    @pytest.mark.skip(reason="parse() method handling is currently commented out in the implementation")
    def test_parse_method(self):
        """Test models with parse method."""
        parse_model = ModelWithParse({"name": "test", "value": 42})
        assert model_to_dict(parse_model) == {"name": "test", "value": 42}

    def test_dict_fallback(self):
        """Test fallback to __dict__."""
        simple_model = SimpleModel("test value")
        assert model_to_dict(simple_model) == {"value": "test value"}

    def test_dict_fallback_exception_handling(self):
        """Test exception handling in dict fallback."""
        # Test with object that has no __dict__ attribute
        model_without_dict = ModelWithoutDict("test value")
        assert model_to_dict(model_without_dict) == {}

        # Test with object that raises exception when accessing __dict__
        class BadModel:
            @property
            def __dict__(self):
                raise AttributeError("No dict for you!")

        bad_model = BadModel()
        assert model_to_dict(bad_model) == {}



"""Common metrics utilities for AgentOps instrumentation.

This module provides utilities for creating and managing standard metrics
across different instrumentations.
"""

from typing import Dict, Any, Optional
from opentelemetry.metrics import Meter, Histogram, Counter
from agentops.semconv import Meters


class StandardMetrics:
    """Factory for creating standard metrics used across instrumentations."""

    @staticmethod
    def create_token_histogram(meter: Meter) -> Histogram:
        """Create a histogram for token usage."""
        return meter.create_histogram(
            name=Meters.LLM_TOKEN_USAGE, unit="token", description="Measures number of input and output tokens used"
        )

    @staticmethod
    def create_duration_histogram(meter: Meter) -> Histogram:
        """Create a histogram for operation duration."""
        return meter.create_histogram(
            name=Meters.LLM_OPERATION_DURATION, unit="s", description="GenAI operation duration"
        )

    @staticmethod
    def create_exception_counter(meter: Meter, name: str = Meters.LLM_COMPLETIONS_EXCEPTIONS) -> Counter:
        """Create a counter for exceptions."""
        return meter.create_counter(
            name=name, unit="time", description="Number of exceptions occurred during operations"
        )

    @staticmethod
    def create_choice_counter(meter: Meter) -> Counter:
        """Create a counter for generation choices."""
        return meter.create_counter(
            name=Meters.LLM_GENERATION_CHOICES,
            unit="choice",
            description="Number of choices returned by completions call",
        )

    @staticmethod
    def create_standard_metrics(meter: Meter) -> Dict[str, Any]:
        """Create a standard set of metrics for LLM operations.

        Returns:
            Dictionary with metric names as keys and metric instances as values
        """
        return {
            "token_histogram": StandardMetrics.create_token_histogram(meter),
            "duration_histogram": StandardMetrics.create_duration_histogram(meter),
            "exception_counter": StandardMetrics.create_exception_counter(meter),
        }


class MetricsRecorder:
    """Utility class for recording metrics in a consistent way."""

    def __init__(self, metrics: Dict[str, Any]):
        self.metrics = metrics

    def record_token_usage(
        self,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        attributes: Optional[Dict[str, Any]] = None,
    ):
        """Record token usage metrics."""
        token_histogram = self.metrics.get("token_histogram")
        if not token_histogram:
            return

        attrs = attributes or {}

        if prompt_tokens is not None:
            token_histogram.record(prompt_tokens, attributes={**attrs, "token.type": "input"})

        if completion_tokens is not None:
            token_histogram.record(completion_tokens, attributes={**attrs, "token.type": "output"})

    def record_duration(self, duration: float, attributes: Optional[Dict[str, Any]] = None):
        """Record operation duration."""
        duration_histogram = self.metrics.get("duration_histogram")
        if duration_histogram:
            duration_histogram.record(duration, attributes=attributes or {})

    def record_exception(self, attributes: Optional[Dict[str, Any]] = None):
        """Record an exception occurrence."""
        exception_counter = self.metrics.get("exception_counter")
        if exception_counter:
            exception_counter.add(1, attributes=attributes or {})

    def record_choices(self, count: int, attributes: Optional[Dict[str, Any]] = None):
        """Record number of choices returned."""
        choice_counter = self.metrics.get("choice_counter")
        if choice_counter:
            choice_counter.add(count, attributes=attributes or {})



"""Span kinds for AgentOps."""

from enum import Enum


class AgentOpsSpanKindValues(Enum):
    """Standard span kind values for AgentOps."""

    WORKFLOW = "workflow"
    SESSION = "session"
    TASK = "task"
    OPERATION = "operation"
    AGENT = "agent"
    TOOL = "tool"
    LLM = "llm"
    TEAM = "team"
    CHAIN = "chain"
    TEXT = "text"
    GUARDRAIL = "guardrail"
    HTTP = "http"
    UNKNOWN = "unknown"


# Legacy SpanKind class for backward compatibility
class SpanKind:
    """Legacy span kind definitions - use AgentOpsSpanKindValues instead."""

    # Agent action kinds
    AGENT_ACTION = "agent.action"  # Agent performing an action
    AGENT_THINKING = "agent.thinking"  # Agent reasoning/planning
    AGENT_DECISION = "agent.decision"  # Agent making a decision

    # LLM interaction kinds
    LLM_CALL = "llm.call"  # LLM API call

    # Workflow kinds
    WORKFLOW_STEP = "workflow.step"  # Step in a workflow
    WORKFLOW = AgentOpsSpanKindValues.WORKFLOW.value
    SESSION = AgentOpsSpanKindValues.SESSION.value
    TASK = AgentOpsSpanKindValues.TASK.value
    OPERATION = AgentOpsSpanKindValues.OPERATION.value
    AGENT = AgentOpsSpanKindValues.AGENT.value
    TOOL = AgentOpsSpanKindValues.TOOL.value
    LLM = AgentOpsSpanKindValues.LLM.value
    TEAM = AgentOpsSpanKindValues.TEAM.value
    UNKNOWN = AgentOpsSpanKindValues.UNKNOWN.value
    CHAIN = AgentOpsSpanKindValues.CHAIN.value
    TEXT = AgentOpsSpanKindValues.TEXT.value
    GUARDRAIL = AgentOpsSpanKindValues.GUARDRAIL.value
    HTTP = AgentOpsSpanKindValues.HTTP.value



"""
Decorators for instrumenting code with AgentOps.
Provides @trace for creating trace-level spans (sessions) and other decorators for nested spans.
"""

from agentops.helpers.deprecation import deprecated
from agentops.sdk.decorators.factory import create_entity_decorator
from agentops.semconv.span_kinds import SpanKind

# Create decorators for specific entity types using the factory
agent = create_entity_decorator(SpanKind.AGENT)
task = create_entity_decorator(SpanKind.TASK)
operation_decorator = create_entity_decorator(SpanKind.OPERATION)
workflow = create_entity_decorator(SpanKind.WORKFLOW)
trace = create_entity_decorator(SpanKind.SESSION)
tool = create_entity_decorator(SpanKind.TOOL)
operation = task
guardrail = create_entity_decorator(SpanKind.GUARDRAIL)
track_endpoint = create_entity_decorator(SpanKind.HTTP)


# For backward compatibility: @session decorator calls @trace decorator
def session(*args, **kwargs):  # noqa: F811
    """@deprecated Use @agentops.trace instead. Wraps the @trace decorator for backward compatibility."""
    # If called as @session or @session(...)
    if not args or not callable(args[0]):  # called with kwargs like @session(name=...)
        return trace(*args, **kwargs)
    else:  # called as @session directly on a function
        return trace(args[0], **kwargs)  # args[0] is the wrapped function


# Apply deprecation decorator to session function
session = deprecated("Use @trace decorator instead.")(session)


# Note: The original `operation = task` was potentially problematic if `operation` was meant to be distinct.
# Using operation_decorator for clarity if a distinct OPERATION kind decorator is needed.
# For now, keeping the alias as it was, assuming it was intentional for `operation` to be `task`.
operation = task

__all__ = [
    "agent",
    "task",
    "workflow",
    "trace",
    "session",
    "operation",
    "tool",
    "guardrail",
    "track_endpoint",
]



"""
AgentOps Validation Module

This module provides functions to validate that spans have been sent to AgentOps
using the public API. This is useful for testing and verification purposes.
"""

import time
import requests
from typing import Optional, Dict, List, Any, Tuple

from agentops.logging import logger
from agentops.exceptions import ApiServerException


class ValidationError(Exception):
    """Raised when span validation fails."""

    pass


def get_jwt_token(api_key: Optional[str] = None) -> str:
    """
    Exchange API key for JWT token.

    Args:
        api_key: Optional API key. If not provided, uses AGENTOPS_API_KEY env var.

    Returns:
        JWT bearer token

    Raises:
        ApiServerException: If token exchange fails
    """
    if api_key is None:
        from agentops import get_client

        client = get_client()
        if client and client.config.api_key:
            api_key = client.config.api_key
        else:
            import os

            api_key = os.getenv("AGENTOPS_API_KEY")
            if not api_key:
                raise ValueError("No API key provided and AGENTOPS_API_KEY environment variable not set")

    try:
        response = requests.post(
            "https://api.agentops.ai/public/v1/auth/access_token", json={"api_key": api_key}, timeout=10
        )
        response.raise_for_status()
        return response.json()["bearer"]
    except requests.exceptions.RequestException as e:
        raise ApiServerException(f"Failed to get JWT token: {e}")


def get_trace_details(trace_id: str, jwt_token: str) -> Dict[str, Any]:
    """
    Get trace details from AgentOps API.

    Args:
        trace_id: The trace ID to query
        jwt_token: JWT authentication token

    Returns:
        Trace details including spans

    Raises:
        ApiServerException: If API request fails
    """
    try:
        response = requests.get(
            f"https://api.agentops.ai/public/v1/traces/{trace_id}",
            headers={"Authorization": f"Bearer {jwt_token}"},
            timeout=10,
        )
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        raise ApiServerException(f"Failed to get trace details: {e}")


def get_trace_metrics(trace_id: str, jwt_token: str) -> Dict[str, Any]:
    """
    Get trace metrics from AgentOps API.

    Args:
        trace_id: The trace ID to query
        jwt_token: JWT authentication token

    Returns:
        Trace metrics including token counts and costs

    Raises:
        ApiServerException: If API request fails
    """
    try:
        response = requests.get(
            f"https://api.agentops.ai/public/v1/traces/{trace_id}/metrics",
            headers={"Authorization": f"Bearer {jwt_token}"},
            timeout=10,
        )
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        raise ApiServerException(f"Failed to get trace metrics: {e}")


def check_llm_spans(spans: List[Dict[str, Any]]) -> Tuple[bool, List[str]]:
    """
    Check if any LLM spans are present in the trace.

    Args:
        spans: List of span dictionaries

    Returns:
        Tuple of (has_llm_spans, llm_span_names)
    """
    llm_spans = []

    for span in spans:
        span_name = span.get("span_name", "unnamed_span")

        # Check span attributes for LLM span kind
        span_attributes = span.get("span_attributes", {})
        is_llm_span = False

        # If we have span_attributes, check them
        if span_attributes:
            # Try different possible structures for span kind
            span_kind = None

            # Structure 1: span_attributes.agentops.span.kind
            if isinstance(span_attributes, dict):
                agentops_attrs = span_attributes.get("agentops", {})
                if isinstance(agentops_attrs, dict):
                    span_info = agentops_attrs.get("span", {})
                    if isinstance(span_info, dict):
                        span_kind = span_info.get("kind", "")

            # Structure 2: Direct in span_attributes
            if not span_kind and isinstance(span_attributes, dict):
                # Try looking for agentops.span.kind as a flattened key
                span_kind = span_attributes.get("agentops.span.kind", "")

            # Structure 3: Look for SpanAttributes.AGENTOPS_SPAN_KIND
            if not span_kind and isinstance(span_attributes, dict):
                from agentops.semconv import SpanAttributes

                span_kind = span_attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND, "")

            # Check if this is an LLM span by span kind
            is_llm_span = span_kind == "llm"

            # Alternative check: Look for gen_ai.prompt or gen_ai.completion attributes
            # These are standard semantic conventions for LLM spans
            if not is_llm_span and isinstance(span_attributes, dict):
                gen_ai_attrs = span_attributes.get("gen_ai", {})
                if isinstance(gen_ai_attrs, dict):
                    # If we have prompt or completion data, it's an LLM span
                    if "prompt" in gen_ai_attrs or "completion" in gen_ai_attrs:
                        is_llm_span = True

            # Check for LLM_REQUEST_TYPE attribute (used by provider instrumentations)
            if not is_llm_span and isinstance(span_attributes, dict):
                from agentops.semconv import SpanAttributes, LLMRequestTypeValues

                # Check for LLM request type - try both gen_ai.* and llm.* prefixes
                # The instrumentation sets gen_ai.* but the API might return llm.*
                llm_request_type = span_attributes.get(SpanAttributes.LLM_REQUEST_TYPE, "")
                if not llm_request_type:
                    # Try the llm.* prefix version
                    llm_request_type = span_attributes.get("llm.request.type", "")

                # Check if it's a chat or completion request (the main LLM types)
                if llm_request_type in [LLMRequestTypeValues.CHAT.value, LLMRequestTypeValues.COMPLETION.value]:
                    is_llm_span = True

        if is_llm_span:
            llm_spans.append(span_name)

    return len(llm_spans) > 0, llm_spans


def validate_trace_spans(
    trace_id: Optional[str] = None,
    trace_context: Optional[Any] = None,
    max_retries: int = 10,
    retry_delay: float = 1.0,
    check_llm: bool = True,
    min_spans: int = 1,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Validate that spans have been sent to AgentOps.

    Args:
        trace_id: Direct trace ID to validate
        trace_context: TraceContext object from start_trace (alternative to trace_id)
        max_retries: Maximum number of retries to wait for spans to appear
        retry_delay: Delay between retries in seconds
        check_llm: Whether to specifically check for LLM spans
        min_spans: Minimum number of spans expected
        api_key: Optional API key (uses environment variable if not provided)

    Returns:
        Dictionary containing validation results and metrics

    Raises:
        ValidationError: If validation fails
        ValueError: If neither trace_id nor trace_context is provided
    """
    # Extract trace ID
    if trace_id is None and trace_context is None:
        # Try to get from current span
        try:
            from opentelemetry.trace import get_current_span

            current_span = get_current_span()
            if current_span and hasattr(current_span, "get_span_context"):
                span_context = current_span.get_span_context()
                if hasattr(span_context, "trace_id") and span_context.trace_id:
                    if isinstance(span_context.trace_id, int):
                        trace_id = format(span_context.trace_id, "032x")
                    else:
                        trace_id = str(span_context.trace_id)
        except ImportError:
            pass

    elif trace_context is not None and trace_id is None:
        # Extract from TraceContext
        if hasattr(trace_context, "span") and trace_context.span:
            span_context = trace_context.span.get_span_context()
            if hasattr(span_context, "trace_id"):
                if isinstance(span_context.trace_id, int):
                    trace_id = format(span_context.trace_id, "032x")
                else:
                    trace_id = str(span_context.trace_id)

    if trace_id is None:
        raise ValueError("No trace ID found. Provide either trace_id or trace_context parameter.")

    # Get JWT token
    jwt_token = get_jwt_token(api_key)

    logger.info(f"Validating spans for trace ID: {trace_id}")

    for attempt in range(max_retries):
        try:
            # Get trace details
            trace_details = get_trace_details(trace_id, jwt_token)
            spans = trace_details.get("spans", [])

            if len(spans) >= min_spans:
                logger.info(f"Found {len(spans)} span(s) in trace")

                # Prepare result
                result = {
                    "trace_id": trace_id,
                    "span_count": len(spans),
                    "spans": spans,
                    "has_llm_spans": False,
                    "llm_span_names": [],
                    "metrics": None,
                }

                # Get metrics first - if we have token usage, we definitely have LLM spans
                try:
                    metrics = get_trace_metrics(trace_id, jwt_token)
                    result["metrics"] = metrics

                    if metrics:
                        logger.info(
                            f"Trace metrics - Total tokens: {metrics.get('total_tokens', 0)}, "
                            f"Cost: ${metrics.get('total_cost', '0.0000')}"
                        )

                        # If we have token usage > 0, we definitely have LLM activity
                        if metrics.get("total_tokens", 0) > 0:
                            result["has_llm_spans"] = True
                            logger.info("LLM activity confirmed via token usage metrics")
                except Exception as e:
                    logger.warning(f"Could not retrieve metrics: {e}")

                # Check for LLM spans if requested and not already confirmed via metrics
                if check_llm and not result["has_llm_spans"]:
                    has_llm_spans, llm_span_names = check_llm_spans(spans)
                    result["has_llm_spans"] = has_llm_spans
                    result["llm_span_names"] = llm_span_names

                    if has_llm_spans:
                        logger.info(f"Found LLM spans: {', '.join(llm_span_names)}")
                    else:
                        logger.warning("No LLM spans found via attribute inspection")

                # Final validation
                if check_llm and not result["has_llm_spans"]:
                    raise ValidationError(
                        f"No LLM activity detected in trace {trace_id}. "
                        f"Found spans: {[s.get('span_name', 'unnamed') for s in spans]}, "
                        f"Token usage: {result.get('metrics', {}).get('total_tokens', 0)}"
                    )

                return result

            else:
                logger.debug(
                    f"Only {len(spans)} spans found, expected at least {min_spans}. "
                    f"Retrying... ({attempt + 1}/{max_retries})"
                )

        except ApiServerException as e:
            logger.debug(f"API error during validation: {e}. Retrying... ({attempt + 1}/{max_retries})")

        if attempt < max_retries - 1:
            time.sleep(retry_delay)

    raise ValidationError(
        f"Validation failed for trace {trace_id} after {max_retries} attempts. "
        f"Expected at least {min_spans} spans"
        + (", including LLM activity" if check_llm else "")
        + ". Please check that tracking is properly configured."
    )


def print_validation_summary(result: Dict[str, Any]) -> None:
    """
    Print a user-friendly summary of validation results.

    Args:
        result: Validation result dictionary from validate_trace_spans
    """
    print("\n" + "=" * 50)
    print("🔍 AgentOps Span Validation Results")
    print("=" * 50)

    print(f"✅ Found {result['span_count']} span(s) in trace")

    if result.get("has_llm_spans"):
        if result.get("llm_span_names"):
            print(f"✅ Found LLM spans: {', '.join(result['llm_span_names'])}")
        else:
            # LLM activity confirmed via metrics
            print("✅ LLM activity confirmed via token usage metrics")
    elif "has_llm_spans" in result:
        print("⚠️  No LLM activity detected")

    if result.get("metrics"):
        metrics = result["metrics"]
        print("\n📊 Trace Metrics:")
        print(f"   - Total tokens: {metrics.get('total_tokens', 0)}")
        print(f"   - Prompt tokens: {metrics.get('prompt_tokens', 0)}")
        print(f"   - Completion tokens: {metrics.get('completion_tokens', 0)}")
        print(f"   - Total cost: ${metrics.get('total_cost', '0.0000')}")

    print("\n✅ Validation successful!")



"""
Unit tests for URL logging functionality and InternalSpanProcessor.
"""

import unittest
from unittest.mock import patch, MagicMock

from opentelemetry.sdk.trace import Span, ReadableSpan

from agentops.sdk.processors import InternalSpanProcessor
from agentops.sdk.core import TraceContext, tracer


class TestURLLogging(unittest.TestCase):
    """Tests for URL logging functionality in global tracer."""

    def setUp(self):
        self.tracing_core = tracer
        # Mock the initialization to avoid actual setup
        self.tracing_core._initialized = True
        self.tracing_core._config = {"project_id": "test_project"}

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    def test_start_trace_logs_url(self, mock_make_span, mock_log_trace_url):
        """Test that start_trace logs the trace URL."""
        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        # Call start_trace
        trace_context = self.tracing_core.start_trace(trace_name="test_trace")

        # Assert that log_trace_url was called with the span and title
        mock_log_trace_url.assert_called_once_with(mock_span, title="test_trace")
        self.assertIsInstance(trace_context, TraceContext)
        self.assertEqual(trace_context.span, mock_span)

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.finalize_span")
    def test_end_trace_logs_url(self, mock_finalize_span, mock_log_trace_url):
        """Test that end_trace logs the trace URL."""
        # Create a mock trace context
        mock_span = MagicMock(spec=Span)
        mock_span.name = "test_trace"
        mock_span.get_span_context.return_value.span_id = 12345
        mock_token = MagicMock()
        trace_context = TraceContext(mock_span, mock_token)

        # Call end_trace
        self.tracing_core.end_trace(trace_context, "Success")

        # Assert that log_trace_url was called with the span and title
        mock_log_trace_url.assert_called_once_with(mock_span, title="test_trace")

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    def test_start_trace_url_logging_failure_does_not_break_trace(self, mock_make_span, mock_log_trace_url):
        """Test that URL logging failure doesn't break trace creation."""
        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        # Make log_trace_url raise an exception
        mock_log_trace_url.side_effect = Exception("URL logging failed")

        # Call start_trace - should not raise exception
        trace_context = self.tracing_core.start_trace(trace_name="test_trace")

        # Assert that trace was still created successfully
        self.assertIsInstance(trace_context, TraceContext)
        self.assertEqual(trace_context.span, mock_span)
        mock_log_trace_url.assert_called_once_with(mock_span, title="test_trace")

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.finalize_span")
    def test_end_trace_url_logging_failure_does_not_break_trace(self, mock_finalize_span, mock_log_trace_url):
        """Test that URL logging failure doesn't break trace ending."""
        # Create a mock trace context
        mock_span = MagicMock(spec=Span)
        mock_span.name = "test_trace"
        mock_span.get_span_context.return_value.span_id = 12345
        mock_token = MagicMock()
        trace_context = TraceContext(mock_span, mock_token)

        # Make log_trace_url raise an exception
        mock_log_trace_url.side_effect = Exception("URL logging failed")

        # Call end_trace - should not raise exception
        self.tracing_core.end_trace(trace_context, "Success")

        # Assert that finalize_span was still called
        mock_finalize_span.assert_called_once()
        mock_log_trace_url.assert_called_once_with(mock_span, title="test_trace")

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    def test_start_trace_with_tags_logs_url(self, mock_make_span, mock_log_trace_url):
        """Test that start_trace with tags logs the trace URL."""
        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        # Call start_trace with tags
        trace_context = self.tracing_core.start_trace(trace_name="tagged_trace", tags=["test", "integration"])

        # Assert that log_trace_url was called with the span and title
        mock_log_trace_url.assert_called_once_with(mock_span, title="tagged_trace")
        self.assertIsInstance(trace_context, TraceContext)


class TestSessionDecoratorURLLogging(unittest.TestCase):
    """Tests for URL logging functionality in session decorators."""

    def setUp(self):
        self.tracing_core = tracer
        # Mock the initialization to avoid actual setup
        self.tracing_core._initialized = True
        self.tracing_core._config = {"project_id": "test_project"}

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    @patch("agentops.sdk.core.tracer.finalize_span")
    def test_session_decorator_logs_url_on_start_and_end(self, mock_finalize_span, mock_make_span, mock_log_trace_url):
        """Test that session decorator logs URLs on both start and end."""
        from agentops.sdk.decorators import session

        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_span.name = "test_function"
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        @session(name="test_session")
        def test_function():
            return "test_result"

        # Call the decorated function
        result = test_function()

        # Assert that log_trace_url was called (start and end)
        # Note: The actual number of calls may vary based on implementation details
        self.assertGreaterEqual(mock_log_trace_url.call_count, 2)
        # Verify that the calls include the expected session name
        call_args_list = [
            call_args[1]["title"] for call_args in mock_log_trace_url.call_args_list if "title" in call_args[1]
        ]
        self.assertIn("test_session", call_args_list)
        self.assertEqual(result, "test_result")

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    @patch("agentops.sdk.core.tracer.finalize_span")
    def test_session_decorator_with_default_name_logs_url(self, mock_finalize_span, mock_make_span, mock_log_trace_url):
        """Test that session decorator with default name logs URLs."""
        from agentops.sdk.decorators import session

        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_span.name = "my_function"
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        @session
        def my_function():
            return "result"

        # Call the decorated function
        result = my_function()

        # Assert that log_trace_url was called with function name as title
        self.assertGreaterEqual(mock_log_trace_url.call_count, 2)
        # Verify that the calls include the expected function name
        call_args_list = [
            call_args[1]["title"] for call_args in mock_log_trace_url.call_args_list if "title" in call_args[1]
        ]
        self.assertIn("my_function", call_args_list)
        self.assertEqual(result, "result")

    @patch("agentops.sdk.core.log_trace_url")
    @patch("agentops.sdk.core.tracer.make_span")
    @patch("agentops.sdk.core.tracer.finalize_span")
    def test_session_decorator_handles_url_logging_failure(
        self, mock_finalize_span, mock_make_span, mock_log_trace_url
    ):
        """Test that session decorator handles URL logging failures gracefully."""
        from agentops.sdk.decorators import session

        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_span.name = "test_function"
        mock_context = MagicMock()
        mock_token = MagicMock()
        mock_span.get_span_context.return_value.span_id = 12345
        mock_make_span.return_value = (mock_span, mock_context, mock_token)

        # Make log_trace_url raise an exception
        mock_log_trace_url.side_effect = Exception("URL logging failed")

        @session(name="failing_session")
        def test_function():
            return "test_result"

        # Call the decorated function - should not raise exception
        result = test_function()

        # Assert that function still executed successfully
        self.assertEqual(result, "test_result")
        # Assert that log_trace_url was called (even though it failed)
        self.assertGreaterEqual(mock_log_trace_url.call_count, 2)


class TestInternalSpanProcessor(unittest.TestCase):
    """Tests for InternalSpanProcessor functionality."""

    def setUp(self):
        self.processor = InternalSpanProcessor()
        # Reset the root span ID before each test
        self.processor._root_span_id = None

    def test_tracks_root_span_on_start(self):
        """Test that the processor tracks the first span as root span."""
        # Create a mock span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = True
        mock_context.span_id = 12345
        mock_span.context = mock_context

        # Call on_start
        self.processor.on_start(mock_span)

        # Assert that root span ID was set
        self.assertEqual(self.processor._root_span_id, 12345)

    def test_ignores_unsampled_spans_on_start(self):
        """Test that unsampled spans are ignored on start."""
        # Create a mock unsampled span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = False
        mock_span.context = mock_context

        # Call on_start
        self.processor.on_start(mock_span)

        # Assert that root span ID was not set
        self.assertIsNone(self.processor._root_span_id)

    def test_only_tracks_first_span_as_root(self):
        """Test that only the first span is tracked as root span."""
        # First span
        mock_span1 = MagicMock(spec=Span)
        mock_context1 = MagicMock()
        mock_context1.trace_flags.sampled = True
        mock_context1.span_id = 12345
        mock_span1.context = mock_context1

        # Second span
        mock_span2 = MagicMock(spec=Span)
        mock_context2 = MagicMock()
        mock_context2.trace_flags.sampled = True
        mock_context2.span_id = 67890
        mock_span2.context = mock_context2

        # Start first span
        self.processor.on_start(mock_span1)
        self.assertEqual(self.processor._root_span_id, 12345)

        # Start second span - should not change root span ID
        self.processor.on_start(mock_span2)
        self.assertEqual(self.processor._root_span_id, 12345)

    @patch("agentops.sdk.processors.upload_logfile")
    def test_uploads_logfile_on_root_span_end(self, mock_upload_logfile):
        """Test that logfile is uploaded when root span ends."""
        # Set up root span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = True
        mock_context.span_id = 12345
        mock_context.trace_id = 98765
        mock_span.context = mock_context

        # Start the span to set it as root
        self.processor.on_start(mock_span)

        # Create readable span for end event
        mock_readable_span = MagicMock(spec=ReadableSpan)
        mock_readable_span.context = mock_context

        # End the span
        self.processor.on_end(mock_readable_span)

        # Assert that upload_logfile was called with trace_id
        mock_upload_logfile.assert_called_once_with(98765)

    @patch("agentops.sdk.processors.upload_logfile")
    def test_does_not_upload_logfile_for_non_root_span(self, mock_upload_logfile):
        """Test that logfile is not uploaded for non-root spans."""
        # Set up root span
        root_span = MagicMock(spec=Span)
        root_context = MagicMock()
        root_context.trace_flags.sampled = True
        root_context.span_id = 12345
        root_span.context = root_context

        # Start root span
        self.processor.on_start(root_span)

        # Create non-root span
        non_root_span = MagicMock(spec=ReadableSpan)
        non_root_context = MagicMock()
        non_root_context.trace_flags.sampled = True
        non_root_context.span_id = 67890  # Different from root
        non_root_span.context = non_root_context

        # End non-root span
        self.processor.on_end(non_root_span)

        # Assert that upload_logfile was not called
        mock_upload_logfile.assert_not_called()

    @patch("agentops.sdk.processors.upload_logfile")
    def test_handles_upload_logfile_error(self, mock_upload_logfile):
        """Test that processor handles upload_logfile errors gracefully."""
        # Set up root span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = True
        mock_context.span_id = 12345
        mock_context.trace_id = 98765
        mock_span.context = mock_context

        # Start the span to set it as root
        self.processor.on_start(mock_span)

        # Make upload_logfile raise an exception
        mock_upload_logfile.side_effect = Exception("Upload failed")

        # Create readable span for end event
        mock_readable_span = MagicMock(spec=ReadableSpan)
        mock_readable_span.context = mock_context

        # End the span - should not raise exception
        self.processor.on_end(mock_readable_span)

        # Assert that upload_logfile was called
        mock_upload_logfile.assert_called_once_with(98765)

    def test_ignores_unsampled_spans_on_end(self):
        """Test that unsampled spans are ignored on end."""
        # Create a mock unsampled span
        mock_span = MagicMock(spec=ReadableSpan)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = False
        mock_span.context = mock_context

        # Call on_end - should not raise exception
        self.processor.on_end(mock_span)

    def test_shutdown_resets_root_span_id(self):
        """Test that shutdown resets the root span ID."""
        # Set up root span
        mock_span = MagicMock(spec=Span)
        mock_context = MagicMock()
        mock_context.trace_flags.sampled = True
        mock_context.span_id = 12345
        mock_span.context = mock_context

        # Start span to set root span ID
        self.processor.on_start(mock_span)
        self.assertEqual(self.processor._root_span_id, 12345)

        # Call shutdown
        self.processor.shutdown()

        # Verify root span ID was reset
        self.assertIsNone(self.processor._root_span_id)

    def test_force_flush_returns_true(self):
        """Test that force_flush returns True."""
        result = self.processor.force_flush()
        self.assertTrue(result)



import asyncio
import inspect
import pytest

from agentops.sdk.decorators.factory import create_entity_decorator
from agentops.semconv import SpanKind
from agentops.semconv.span_attributes import SpanAttributes
from tests.unit.sdk.instrumentation_tester import InstrumentationTester
from agentops.sdk.core import tracer


class TestFactoryModule:
    """Comprehensive tests for the factory.py module functionality."""

    def test_create_entity_decorator_factory_function(self):
        """Test that create_entity_decorator returns a callable decorator."""
        decorator = create_entity_decorator("test_kind")
        assert callable(decorator)

        # Test that it can be used as a decorator
        @decorator
        def test_function():
            return "test"

        assert test_function() == "test"

    def test_decorator_with_parameters(self):
        """Test decorator with explicit parameters."""
        decorator = create_entity_decorator("test_kind")

        @decorator(name="custom_name", version="1.0", tags=["tag1", "tag2"])
        def test_function():
            return "test"

        assert test_function() == "test"

    def test_decorator_partial_application(self):
        """Test that decorator can be partially applied."""
        decorator = create_entity_decorator("test_kind")
        partial_decorator = decorator(name="partial_name", version="2.0")

        @partial_decorator
        def test_function():
            return "test"

        assert test_function() == "test"

    def test_class_decoration_basic(self, instrumentation: InstrumentationTester):
        """Test basic class decoration functionality."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self, value=42):
                self.value = value

        # Test instantiation
        instance = TestClass(100)
        assert instance.value == 100

        # Note: The current factory implementation has a bug where class decoration
        # creates spans but doesn't properly end them, so no spans are recorded
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 0

    def test_class_decoration_with_parameters(self, instrumentation: InstrumentationTester):
        """Test class decoration with explicit parameters."""
        decorator = create_entity_decorator("test_kind")

        @decorator(name="CustomClass", version="1.0", tags={"env": "test"})
        class TestClass:
            def __init__(self, value=42):
                self.value = value

        instance = TestClass(100)
        assert instance.value == 100

        # Note: The current factory implementation has a bug where class decoration
        # creates spans but doesn't properly end them, so no spans are recorded
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 0

    def test_class_metadata_preservation(self):
        """Test that class metadata is preserved after decoration."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            """Test class docstring."""

            def __init__(self):
                pass

        assert TestClass.__name__ == "TestClass"
        # The qualname will include the test function context, which is expected
        assert "TestClass" in TestClass.__qualname__
        assert TestClass.__module__ == TestClass.__module__  # Should be preserved
        assert TestClass.__doc__ == "Test class docstring."

    def test_async_context_manager_normal_flow(self, instrumentation: InstrumentationTester):
        """Test async context manager with normal flow."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self):
                self.value = 42

        async def test_async_context():
            async with TestClass() as instance:
                assert instance.value == 42
                assert hasattr(instance, "_agentops_active_span")
                assert instance._agentops_active_span is not None
                return "success"

        result = asyncio.run(test_async_context())
        assert result == "success"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_async_context_manager_exception_flow(self, instrumentation: InstrumentationTester):
        """Test async context manager with exception flow."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self):
                self.value = 42

        async def test_async_context_with_exception():
            try:
                async with TestClass() as instance:
                    assert instance.value == 42
                    raise ValueError("Test exception")
            except ValueError:
                return "exception_handled"

        result = asyncio.run(test_async_context_with_exception())
        assert result == "exception_handled"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_async_context_manager_reuse(self, instrumentation: InstrumentationTester):
        """Test that async context manager can be reused."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self):
                self.value = 42

        async def test_reuse():
            instance = TestClass()

            # First use
            async with instance as inst1:
                assert inst1.value == 42

            # Second use - should work with existing span
            async with instance as inst2:
                assert inst2.value == 42
                assert inst2 is instance

        asyncio.run(test_reuse())

        spans = instrumentation.get_finished_spans()
        # The current implementation creates a span for __init__ and another for the async context
        assert len(spans) == 2

    def test_sync_function_decoration(self, instrumentation: InstrumentationTester):
        """Test synchronous function decoration."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_function(x, y=10):
            return x + y

        result = test_function(5, y=15)
        assert result == 20

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_function.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_async_function_decoration(self, instrumentation: InstrumentationTester):
        """Test asynchronous function decoration."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_function(x, y=10):
            await asyncio.sleep(0.01)
            return x + y

        result = asyncio.run(test_async_function(5, y=15))
        assert result == 20

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_async_function.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_generator_function_decoration(self, instrumentation: InstrumentationTester):
        """Test generator function decoration."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_generator(count):
            for i in range(count):
                yield f"item_{i}"

        results = list(test_generator(3))
        assert results == ["item_0", "item_1", "item_2"]

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_generator.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_async_generator_function_decoration(self, instrumentation: InstrumentationTester):
        """Test async generator function decoration."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_generator(count):
            for i in range(count):
                await asyncio.sleep(0.01)
                yield f"async_item_{i}"

        async def collect_results():
            results = []
            async for item in test_async_generator(3):
                results.append(item)
            return results

        results = asyncio.run(collect_results())
        assert results == ["async_item_0", "async_item_1", "async_item_2"]

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_async_generator.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_session_entity_kind_sync(self, instrumentation: InstrumentationTester):
        """Test SESSION entity kind with sync function."""
        decorator = create_entity_decorator("session")

        @decorator
        def test_session_function():
            return "session_result"

        result = test_session_function()
        assert result == "session_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_session_function.session"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "session"

    def test_session_entity_kind_async(self, instrumentation: InstrumentationTester):
        """Test SESSION entity kind with async function."""
        decorator = create_entity_decorator("session")

        @decorator
        async def test_session_async_function():
            await asyncio.sleep(0.01)
            return "session_async_result"

        result = asyncio.run(test_session_async_function())
        assert result == "session_async_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_session_async_function.session"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "session"

    def test_session_entity_kind_generator_warning(self, caplog, instrumentation: InstrumentationTester):
        """Test that SESSION entity kind logs warning for generators."""
        # TODO: This test should assert that a warning is logged, but logger capture is complex due to custom logger setup.
        # For now, we only assert the correct span behavior.
        decorator = create_entity_decorator("session")

        @decorator
        def test_session_generator():
            yield "session_generator_item"

        # The decorator should return a generator, not None
        # For session decorators, the generator logic falls through to the regular generator handling
        generator = test_session_generator()
        results = list(generator)
        assert results == ["session_generator_item"]

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_tool_entity_kind_with_cost(self, instrumentation: InstrumentationTester):
        """Test tool entity kind with cost parameter."""
        decorator = create_entity_decorator("tool")

        @decorator(cost=0.05)
        def test_tool_function():
            return "tool_result"

        result = test_tool_function()
        assert result == "tool_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_tool_function.tool"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "tool"
        assert span.attributes.get("gen_ai.usage.total_cost") == 0.05

    def test_guardrail_entity_kind_with_spec(self, instrumentation: InstrumentationTester):
        """Test guardrail entity kind with spec parameter."""
        decorator = create_entity_decorator("guardrail")

        @decorator(spec="input")
        def test_guardrail_function():
            return "guardrail_result"

        result = test_guardrail_function()
        assert result == "guardrail_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_guardrail_function.guardrail"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "guardrail"
        assert span.attributes.get("agentops.guardrail.spec") == "input"

    def test_guardrail_entity_kind_with_output_spec(self, instrumentation: InstrumentationTester):
        """Test guardrail entity kind with output spec parameter."""
        decorator = create_entity_decorator("guardrail")

        @decorator(spec="output")
        def test_guardrail_output_function():
            return "guardrail_output_result"

        result = test_guardrail_output_function()
        assert result == "guardrail_output_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_guardrail_output_function.guardrail"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "guardrail"
        assert span.attributes.get("agentops.guardrail.spec") == "output"

    def test_guardrail_entity_kind_with_invalid_spec(self, instrumentation: InstrumentationTester):
        """Test guardrail entity kind with invalid spec parameter."""
        decorator = create_entity_decorator("guardrail")

        @decorator(spec="invalid")
        def test_guardrail_invalid_function():
            return "guardrail_invalid_result"

        result = test_guardrail_invalid_function()
        assert result == "guardrail_invalid_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_guardrail_invalid_function.guardrail"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "guardrail"
        # Should not have the spec attribute for invalid spec
        assert "agentops.decorator.guardrail.spec" not in span.attributes

    def test_tags_parameter_list(self, instrumentation: InstrumentationTester):
        """Test tags parameter with list."""
        decorator = create_entity_decorator("test_kind")

        @decorator(tags=["tag1", "tag2", "tag3"])
        def test_function_with_tags():
            return "tagged_result"

        result = test_function_with_tags()
        assert result == "tagged_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_function_with_tags.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"
        # Tags should be recorded in the span attributes

    def test_tags_parameter_dict(self, instrumentation: InstrumentationTester):
        """Test tags parameter with dictionary."""
        decorator = create_entity_decorator("test_kind")

        @decorator(tags={"env": "test", "version": "1.0"})
        def test_function_with_dict_tags():
            return "dict_tagged_result"

        result = test_function_with_dict_tags()
        assert result == "dict_tagged_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_function_with_dict_tags.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_version_parameter(self, instrumentation: InstrumentationTester):
        """Test version parameter."""
        decorator = create_entity_decorator("test_kind")

        @decorator(version="2.1.0")
        def test_function_with_version():
            return "versioned_result"

        result = test_function_with_version()
        assert result == "versioned_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_function_with_version.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_function_with_exception(self, instrumentation: InstrumentationTester):
        """Test function decoration with exception handling."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_function_with_exception():
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"):
            test_function_with_exception()

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_function_with_exception.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_async_function_with_exception(self, instrumentation: InstrumentationTester):
        """Test async function decoration with exception handling."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_function_with_exception():
            await asyncio.sleep(0.01)
            raise RuntimeError("Async test exception")

        with pytest.raises(RuntimeError, match="Async test exception"):
            asyncio.run(test_async_function_with_exception())

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_async_function_with_exception.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_class_init_with_exception(self, instrumentation: InstrumentationTester):
        """Test class decoration with exception in __init__."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClassWithException:
            def __init__(self, should_raise=False):
                if should_raise:
                    raise ValueError("Init exception")
                self.value = 42

        # Normal instantiation
        instance = TestClassWithException(should_raise=False)
        assert instance.value == 42

        # Exception during instantiation
        with pytest.raises(ValueError, match="Init exception"):
            TestClassWithException(should_raise=True)

        spans = instrumentation.get_finished_spans()
        # Only one span should be created (for the successful instantiation)
        # The failed instantiation doesn't create a span because the exception is raised before span creation
        assert len(spans) == 1

    def test_tracer_not_initialized(self, instrumentation: InstrumentationTester):
        """Test behavior when tracer is not initialized."""
        # We can't directly set tracer.initialized as it's a read-only property
        # Instead, we'll test that the decorator works when tracer is not initialized
        # by temporarily mocking the tracer.initialized property

        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_function_no_tracer():
            return "no_tracer_result"

        # This should work normally since tracer is initialized in tests
        result = test_function_no_tracer()
        assert result == "no_tracer_result"

        # Should create spans normally
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_complex_parameter_combination(self, instrumentation: InstrumentationTester):
        """Test decorator with all parameters combined."""
        decorator = create_entity_decorator("tool")

        @decorator(
            name="complex_function", version="3.0.0", tags={"env": "test", "component": "test"}, cost=0.1, spec="input"
        )
        def test_complex_function(x, y):
            return x * y

        result = test_complex_function(5, 6)
        assert result == 30

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "complex_function.tool"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "tool"
        assert span.attributes.get("gen_ai.usage.total_cost") == 0.1

    def test_method_decoration(self, instrumentation: InstrumentationTester):
        """Test decoration of class methods."""
        decorator = create_entity_decorator("test_kind")

        class TestClass:
            def __init__(self):
                self.value = 0

            @decorator
            def test_method(self, increment):
                self.value += increment
                return self.value

        instance = TestClass()
        result = instance.test_method(5)
        assert result == 5
        assert instance.value == 5

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_method.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_static_method_decoration(self, instrumentation: InstrumentationTester):
        """Test decoration of static methods."""
        decorator = create_entity_decorator("test_kind")

        class TestClass:
            @staticmethod
            @decorator
            def test_static_method(x, y):
                return x + y

        result = TestClass.test_static_method(3, 4)
        assert result == 7

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_static_method.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_class_method_decoration(self, instrumentation: InstrumentationTester):
        """Test decoration of class methods."""
        decorator = create_entity_decorator("test_kind")

        class TestClass:
            class_value = 100

            @classmethod
            @decorator
            def test_class_method(cls, increment):
                cls.class_value += increment
                return cls.class_value

        result = TestClass.test_class_method(50)
        assert result == 150
        assert TestClass.class_value == 150

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_class_method.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_nested_decorators(self, instrumentation: InstrumentationTester):
        """Test multiple decorators applied to the same function."""
        decorator1 = create_entity_decorator("kind1")
        decorator2 = create_entity_decorator("kind2")

        @decorator1
        @decorator2
        def test_nested_function():
            return "nested_result"

        result = test_nested_function()
        assert result == "nested_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 2  # Should create spans for both decorators

        # Check that both spans were created with correct names
        span_names = [span.name for span in spans]
        assert "test_nested_function.kind2" in span_names
        assert "test_nested_function.kind1" in span_names

        span_kinds = [span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) for span in spans]
        assert "kind1" in span_kinds
        assert "kind2" in span_kinds

    def test_decorator_with_lambda(self, instrumentation: InstrumentationTester):
        """Test decorator with lambda function."""
        decorator = create_entity_decorator("test_kind")

        test_lambda = decorator(lambda x: x * 2)

        result = test_lambda(5)
        assert result == 10

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_builtin_function(self, instrumentation: InstrumentationTester):
        """Test decorator with built-in function (should work but may not create spans)."""
        decorator = create_entity_decorator("test_kind")

        # This should not raise an error, but may not create spans due to built-in function limitations
        decorated_len = decorator(len)

        result = decorated_len([1, 2, 3, 4, 5])
        assert result == 5

        # Built-in functions may not be instrumented the same way
        _ = instrumentation.get_finished_spans()
        # The behavior may vary depending on the implementation

    def test_decorator_with_coroutine_function(self, instrumentation: InstrumentationTester):
        """Test decorator with coroutine function."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_coroutine():
            await asyncio.sleep(0.01)
            return "coroutine_result"

        # Test that it's actually a coroutine function
        assert asyncio.iscoroutinefunction(test_coroutine)

        result = asyncio.run(test_coroutine())
        assert result == "coroutine_result"

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_coroutine.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_async_generator_function(self, instrumentation: InstrumentationTester):
        """Test decorator with async generator function."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_gen():
            for i in range(3):
                await asyncio.sleep(0.01)
                yield f"async_gen_item_{i}"

        # Test that it's actually an async generator function
        assert inspect.isasyncgenfunction(test_async_gen)

        async def collect_async_gen():
            results = []
            async for item in test_async_gen():
                results.append(item)
            return results

        results = asyncio.run(collect_async_gen())
        assert results == ["async_gen_item_0", "async_gen_item_1", "async_gen_item_2"]

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_async_gen.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_generator_function(self, instrumentation: InstrumentationTester):
        """Test decorator with generator function."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_gen():
            for i in range(3):
                yield f"gen_item_{i}"

        # Test that it's actually a generator function
        assert inspect.isgeneratorfunction(test_gen)

        results = list(test_gen())
        assert results == ["gen_item_0", "gen_item_1", "gen_item_2"]

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_gen.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_kwargs_only_function(self, instrumentation: InstrumentationTester):
        """Test decorator with function that only accepts kwargs."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_kwargs_only(**kwargs):
            return sum(kwargs.values())

        result = test_kwargs_only(a=1, b=2, c=3)
        assert result == 6

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_kwargs_only.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_args_only_function(self, instrumentation: InstrumentationTester):
        """Test decorator with function that only accepts args."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_args_only(*args):
            return sum(args)

        result = test_args_only(1, 2, 3, 4, 5)
        assert result == 15

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_args_only.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_decorator_with_mixed_args_function(self, instrumentation: InstrumentationTester):
        """Test decorator with function that accepts both args and kwargs."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_mixed_args(x, y, *args, **kwargs):
            return x + y + sum(args) + sum(kwargs.values())

        result = test_mixed_args(1, 2, 3, 4, a=5, b=6)
        assert result == 21  # 1 + 2 + 3 + 4 + 5 + 6

        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "test_mixed_args.test_kind"
        assert span.attributes.get(SpanAttributes.AGENTOPS_SPAN_KIND) == "test_kind"

    def test_class_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording class input fails."""
        decorator = create_entity_decorator("test_kind")

        # Create a class that will cause _record_entity_input to fail
        @decorator
        class TestClass:
            def __init__(self, value=42):
                # Create an object that will cause serialization to fail
                self.value = value
                self.bad_object = object()  # This will cause serialization issues

        # The exception should be caught and logged
        instance = TestClass(100)
        assert instance.value == 100
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_class_output_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording class output fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self):
                self.value = 42
                # Create an object that will cause serialization to fail
                self.bad_object = object()

        async def test_async_context():
            async with TestClass():
                return "success"

        result = asyncio.run(test_async_context())
        assert result == "success"
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_generator_implementation(self, instrumentation: InstrumentationTester, caplog):
        """Test the session generator implementation that was previously not implemented."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_generator():
            yield 1
            yield 2
            yield 3

        results = list(test_session_generator())
        assert results == [1, 2, 3]

        # The warning should be logged, but the exact message might vary
        # Just verify that the function works and creates spans
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_session_async_generator_implementation(self, instrumentation: InstrumentationTester, caplog):
        """Test the session async generator implementation."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_generator():
            yield 1
            yield 2
            yield 3

        async def collect_results():
            results = []
            async for item in test_session_async_generator():
                results.append(item)
            return results

        results = asyncio.run(collect_results())
        assert results == [1, 2, 3]

        # The warning should be logged, but the exact message might vary
        # Just verify that the function works and creates spans
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_session_generator_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session generator input fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_generator():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            yield 1
            yield 2

        results = list(test_session_generator())
        assert results == [1, 2]
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_async_generator_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session async generator input fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_generator():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            yield 1
            yield 2

        async def collect_results():
            results = []
            async for item in test_session_async_generator():
                results.append(item)
            return results

        results = asyncio.run(collect_results())
        assert results == [1, 2]
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_async_trace_start_failure(self, instrumentation: InstrumentationTester, caplog):
        """Test handling when trace start fails for session async function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        # Mock tracer.start_trace to return None
        with pytest.MonkeyPatch().context() as m:
            m.setattr(tracer, "start_trace", lambda *args, **kwargs: None)

            @decorator
            async def test_session_async_function():
                return "success"

            result = asyncio.run(test_session_async_function())
            assert result == "success"
            # The error message should be logged, but the exact format might vary
            # Just verify that the function works when trace start fails

    def test_session_async_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session async input fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_function():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            return "success"

        result = asyncio.run(test_session_async_function())
        assert result == "success"
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_async_output_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session async output fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_function():
            # Return an object that will cause serialization to fail
            return object()

        result = asyncio.run(test_session_async_function())
        assert result is not None
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_async_exception_handling(self, instrumentation: InstrumentationTester):
        """Test exception handling in session async function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_function():
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"):
            asyncio.run(test_session_async_function())

        # Should end trace with "Indeterminate" state
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_session_async_finally_block(self, instrumentation: InstrumentationTester, caplog):
        """Test finally block handling in session async function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        async def test_session_async_function():
            return "success"

        result = asyncio.run(test_session_async_function())
        assert result == "success"

        # Should not log warning about trace not being ended since it was ended properly
        assert "not explicitly ended" not in caplog.text

    def test_session_sync_trace_start_failure(self, instrumentation: InstrumentationTester, caplog):
        """Test handling when trace start fails for session sync function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        # Mock tracer.start_trace to return None
        with pytest.MonkeyPatch().context() as m:
            m.setattr(tracer, "start_trace", lambda *args, **kwargs: None)

            @decorator
            def test_session_sync_function():
                return "success"

            result = test_session_sync_function()
            assert result == "success"
            # The error message should be logged, but the exact format might vary
            # Just verify that the function works when trace start fails

    def test_session_sync_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session sync input fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_sync_function():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            return "success"

        result = test_session_sync_function()
        assert result == "success"
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_sync_output_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording session sync output fails."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_sync_function():
            # Return an object that will cause serialization to fail
            return object()

        result = test_session_sync_function()
        assert result is not None
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_session_sync_exception_handling(self, instrumentation: InstrumentationTester):
        """Test exception handling in session sync function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_sync_function():
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"):
            test_session_sync_function()

        # Should end trace with "Indeterminate" state
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_session_sync_finally_block(self, instrumentation: InstrumentationTester, caplog):
        """Test finally block handling in session sync function."""
        decorator = create_entity_decorator(SpanKind.SESSION)

        @decorator
        def test_session_sync_function():
            return "success"

        result = test_session_sync_function()
        assert result == "success"

        # Should not log warning about trace not being ended since it was ended properly
        assert "not explicitly ended" not in caplog.text

    def test_generator_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording generator input fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_generator():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            yield 1
            yield 2

        results = list(test_generator())
        assert results == [1, 2]
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_async_generator_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording async generator input fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_generator():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            yield 1
            yield 2

        async def collect_results():
            results = []
            async for item in test_async_generator():
                results.append(item)
            return results

        results = asyncio.run(collect_results())
        assert results == [1, 2]
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_async_function_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording async function input fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_function():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            return "success"

        result = asyncio.run(test_async_function())
        assert result == "success"
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_async_function_output_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording async function output fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_function():
            # Return an object that will cause serialization to fail
            return object()

        result = asyncio.run(test_async_function())
        assert result is not None
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_async_function_execution_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling in async function execution."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        async def test_async_function():
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"):
            asyncio.run(test_async_function())

        # The error should be logged, but the exact message might vary
        # Just verify that the exception is handled properly
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_sync_function_input_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording sync function input fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_sync_function():
            # Create an object that will cause serialization to fail
            _ = object()  # This will cause serialization issues
            return "success"

        result = test_sync_function()
        assert result == "success"
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_sync_function_output_recording_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling when recording sync function output fails."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_sync_function():
            # Return an object that will cause serialization to fail
            return object()

        result = test_sync_function()
        assert result is not None
        # Note: The actual exception might not be logged in the current implementation
        # but the coverage will show if the exception handling path was executed

    def test_sync_function_execution_exception(self, instrumentation: InstrumentationTester, caplog):
        """Test exception handling in sync function execution."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        def test_sync_function():
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"):
            test_sync_function()

        # The error should be logged, but the exact message might vary
        # Just verify that the exception is handled properly
        spans = instrumentation.get_finished_spans()
        assert len(spans) == 1

    def test_class_del_method_coverage(self, instrumentation: InstrumentationTester):
        """Test that __del__ method is called when object is garbage collected."""
        decorator = create_entity_decorator("test_kind")

        @decorator
        class TestClass:
            def __init__(self):
                self.value = 42

        # Create instance and let it go out of scope to trigger __del__
        def create_and_destroy():
            instance = TestClass()
            assert instance.value == 42
            # The __del__ method should be called when instance goes out of scope

        create_and_destroy()

        # Force garbage collection to trigger __del__
        import gc

        gc.collect()

        # The __del__ method should have been called, but we can't easily test this
        # since it's called during garbage collection. The coverage will show if the
        # lines were executed.



from unittest.mock import patch, MagicMock
import agentops
import threading


def test_get_client_singleton():
    # Should always return the same instance
    c1 = agentops.get_client()
    c2 = agentops.get_client()
    assert c1 is c2


def test_get_client_thread_safety():
    # Should not create multiple clients in threads
    results = []

    def worker():
        results.append(agentops.get_client())

    threads = [threading.Thread(target=worker) for _ in range(5)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert all(r is results[0] for r in results)


def test_init_merges_tags(monkeypatch):
    with patch("agentops.get_client") as mock_get_client:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        agentops.init(tags=["a"], default_tags=["b"])  # Should merge
        assert {"a", "b"}.issubset(set(mock_client.init.call_args[1]["default_tags"]))


def test_init_warns_on_deprecated_tags(monkeypatch):
    with patch("agentops.get_client") as mock_get_client, patch("agentops.warn_deprecated_param") as mock_warn:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        agentops.init(tags=["a"])
        mock_warn.assert_called_once_with("tags", "default_tags")


def test_init_jupyter_detection(monkeypatch):
    with patch("agentops.get_client") as mock_get_client:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        # Simulate Jupyter by patching get_ipython
        import builtins

        builtins.get_ipython = lambda: type("Z", (), {"__name__": "ZMQInteractiveShell"})()
        agentops.init()
        del builtins.get_ipython


def test_init_jupyter_detection_nameerror():
    with patch("agentops.get_client") as mock_get_client:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        # Simulate NameError when get_ipython() is called
        import builtins

        original_get_ipython = getattr(builtins, "get_ipython", None)
        builtins.get_ipython = lambda: None  # This will cause NameError
        try:
            agentops.init()
        except NameError:
            pass  # Expected
        finally:
            if original_get_ipython:
                builtins.get_ipython = original_get_ipython
            else:
                delattr(builtins, "get_ipython")


def test_configure_valid_and_invalid_params():
    with patch("agentops.get_client") as mock_get_client, patch("agentops.logger") as mock_logger:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        # Valid param
        agentops.configure(api_key="foo")
        mock_client.configure.assert_called_with(api_key="foo")
        # Invalid param
        agentops.configure(bad_param=123)
        mock_logger.warning.assert_any_call("Invalid configuration parameters: {'bad_param'}")


def test_record_sets_end_timestamp():
    class Dummy:
        end_timestamp = None

    with patch("agentops.helpers.time.get_ISO_time", return_value="now"):
        d = Dummy()
        agentops.record(d)
        assert d.end_timestamp == "now"


def test_record_no_end_timestamp():
    class Dummy:
        pass

    d = Dummy()
    assert agentops.record(d) is d


def test_update_trace_metadata_success():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        result = agentops.update_trace_metadata({"foo": "bar"})
        assert result is True
        mock_span.set_attribute.assert_called()


def test_update_trace_metadata_no_active_span():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span", return_value=None):
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        assert not agentops.update_trace_metadata({"foo": "bar"})


def test_update_trace_metadata_not_recording():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = False
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        assert not agentops.update_trace_metadata({"foo": "bar"})


def test_update_trace_metadata_invalid_type():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        # Value is a dict, which is not allowed
        assert not agentops.update_trace_metadata({"foo": {"bar": 1}})


def test_update_trace_metadata_list_type():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        # List of valid types
        assert agentops.update_trace_metadata({"foo": [1, 2, 3]})
        mock_span.set_attribute.assert_called()


def test_update_trace_metadata_extract_key_single_part():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Test with a semantic convention that has only one part (len < 2)
        with patch("agentops.AgentAttributes") as mock_attrs:
            mock_attrs.__dict__ = {"SINGLE": "single_value"}
            result = agentops.update_trace_metadata({"single_value": "test"})
            assert result is True


def test_update_trace_metadata_skip_gen_ai_attributes():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Test that gen_ai attributes are skipped
        with patch("agentops.AgentAttributes") as mock_attrs:
            mock_attrs.__dict__ = {"GEN_AI_ATTR": "gen_ai.something"}
            agentops.update_trace_metadata({"gen_ai.something": "test"})
            # Should still work but skip the gen_ai attribute


def test_update_trace_metadata_trace_id_conversion_error():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "child_span"  # Not a session span
        mock_span.get_span_context.return_value.trace_id = "invalid_hex"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {"trace1": MagicMock()}
        mock_tracer.initialized = True

        # This should handle the ValueError from int("invalid_hex", 16)
        agentops.update_trace_metadata({"foo": "bar"})
        # The function should handle the error gracefully


def test_update_trace_metadata_no_active_traces():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span", return_value=None),
        patch("agentops.logger") as mock_logger,
    ):
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        result = agentops.update_trace_metadata({"foo": "bar"})
        assert result is False
        mock_logger.warning.assert_called_with("No active trace found. Cannot update metadata.")


def test_update_trace_metadata_span_not_recording():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = False
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        result = agentops.update_trace_metadata({"foo": "bar"})
        assert result is False
        mock_logger.warning.assert_called_with("No active trace found. Cannot update metadata.")


def test_update_trace_metadata_list_invalid_types():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # List with invalid types (dict)
        agentops.update_trace_metadata({"foo": [{"invalid": "type"}]})
        mock_logger.warning.assert_called_with("No valid metadata attributes were updated")


def test_update_trace_metadata_invalid_value_type():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Invalid value type (dict)
        agentops.update_trace_metadata({"foo": {"invalid": "type"}})
        mock_logger.warning.assert_called_with("No valid metadata attributes were updated")


def test_update_trace_metadata_semantic_convention_mapping():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Test semantic convention mapping
        with patch("agentops.AgentAttributes") as mock_attrs:
            mock_attrs.__dict__ = {"TEST_ATTR": "agent.test_attribute"}
            agentops.update_trace_metadata({"agent_test_attribute": "test"})
            mock_logger.debug.assert_called_with("Successfully updated 1 metadata attributes on trace")


def test_update_trace_metadata_exception_handling():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_span.set_attribute.side_effect = Exception("Test error")
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        agentops.update_trace_metadata({"foo": "bar"})
        mock_logger.error.assert_called_with("Error updating trace metadata: Test error")


def test_update_trace_metadata_no_valid_attributes():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span") as mock_get_span,
        patch("agentops.logger") as mock_logger,
    ):
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # All values are None
        agentops.update_trace_metadata({"foo": None, "bar": None})
        mock_logger.warning.assert_called_with("No valid metadata attributes were updated")


def test_start_trace_auto_init_failure():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.init") as mock_init,
        patch("agentops.logger") as mock_logger,
    ):
        mock_tracer.initialized = False
        mock_init.side_effect = Exception("Init failed")

        agentops.start_trace("test")
        mock_logger.error.assert_called_with(
            "SDK auto-initialization failed during start_trace: Init failed. Cannot start trace."
        )


def test_start_trace_auto_init_still_not_initialized():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.init") as _,
        patch("agentops.logger") as mock_logger,
    ):
        mock_tracer.initialized = False

        agentops.start_trace("test")
        mock_logger.error.assert_called_with("SDK initialization failed. Cannot start trace.")


def test_end_trace_not_initialized():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.logger") as mock_logger:
        mock_tracer.initialized = False
        agentops.end_trace()
        mock_logger.warning.assert_called_with("AgentOps SDK not initialized. Cannot end trace.")


def test_update_trace_metadata_not_initialized():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.logger") as mock_logger:
        mock_tracer.initialized = False
        agentops.update_trace_metadata({"foo": "bar"})
        mock_logger.warning.assert_called_with("AgentOps SDK not initialized. Cannot update trace metadata.")


def test_all_exports_importable():
    # Just import all symbols to ensure they're present
    from agentops import (
        init,
        configure,
    )

    assert callable(init)
    assert callable(configure)


def test_update_trace_metadata_use_current_span_when_no_parent_found():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "child_span"  # Not a session span
        mock_span.get_span_context.return_value.trace_id = 12345
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {"trace1": MagicMock()}
        mock_tracer.initialized = True

        # When no parent trace is found, should use current span
        agentops.update_trace_metadata({"foo": "bar"})
        # The function should work with current span


def test_update_trace_metadata_use_current_span_when_no_active_traces():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "child_span"  # Not a session span
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # When no active traces, should use current span
        agentops.update_trace_metadata({"foo": "bar"})
        # The function should work with current span


def test_update_trace_metadata_use_most_recent_trace():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span", return_value=None),
        patch("agentops.logger") as mock_logger,
    ):
        mock_trace_context = MagicMock()
        mock_trace_context.span = MagicMock()
        mock_trace_context.span.is_recording.return_value = True
        mock_tracer.get_active_traces.return_value = {"trace1": mock_trace_context}
        mock_tracer.initialized = True

        agentops.update_trace_metadata({"foo": "bar"})
        mock_logger.debug.assert_called_with("Successfully updated 1 metadata attributes on trace")


def test_end_trace_with_trace_context():
    with patch("agentops.tracer") as mock_tracer:
        mock_tracer.initialized = True
        mock_trace_context = MagicMock()
        agentops.end_trace(mock_trace_context, "Error")
        mock_tracer.end_trace.assert_called_with(trace_context=mock_trace_context, end_state="Error")


def test_init_jupyter_detection_actual_nameerror():
    with patch("agentops.get_client") as mock_get_client:
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client
        # Actually remove get_ipython to trigger NameError
        import builtins

        original_get_ipython = getattr(builtins, "get_ipython", None)
        if hasattr(builtins, "get_ipython"):
            delattr(builtins, "get_ipython")
        try:
            agentops.init()
        finally:
            if original_get_ipython:
                builtins.get_ipython = original_get_ipython


def test_end_trace_with_default_state():
    with patch("agentops.tracer") as mock_tracer:
        mock_tracer.initialized = True
        from agentops import TraceState

        agentops.end_trace()  # Should use default TraceState.SUCCESS
        mock_tracer.end_trace.assert_called_with(trace_context=None, end_state=TraceState.SUCCESS)


def test_update_trace_metadata_extract_key_single_part_actual():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Test with a semantic convention that has only one part (len < 2)
        with patch("agentops.AgentAttributes") as mock_attrs:
            mock_attrs.__dict__ = {"SINGLE": "single"}
            agentops.update_trace_metadata({"single": "test"})
            # The function should handle single-part attributes


def test_update_trace_metadata_skip_gen_ai_attributes_actual():
    with patch("agentops.tracer") as mock_tracer, patch("agentops.get_current_span") as mock_get_span:
        mock_span = MagicMock()
        mock_span.is_recording.return_value = True
        mock_span.name = "foo.SESSION"
        mock_get_span.return_value = mock_span
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True

        # Test that gen_ai attributes are actually skipped in the mapping
        with patch("agentops.AgentAttributes") as mock_attrs:
            mock_attrs.__dict__ = {"GEN_AI_ATTR": "gen_ai.something"}
            agentops.update_trace_metadata({"gen_ai.something": "test"})
            # Should still work but the gen_ai attribute should be skipped in mapping


def test_update_trace_metadata_no_active_traces_actual():
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.get_current_span", return_value=None),
        patch("agentops.logger") as mock_logger,
    ):
        mock_tracer.get_active_traces.return_value = {}
        mock_tracer.initialized = True
        agentops.update_trace_metadata({"foo": "bar"})
        mock_logger.warning.assert_called_with("No active trace found. Cannot update metadata.")



"""
API client versions package.

This package contains client implementations for different API versions.
"""

from agentops.client.api.versions.v3 import V3Client
from agentops.client.api.versions.v4 import V4Client

__all__ = ["V3Client", "V4Client"]



"""
V3 API client for the AgentOps API.

This module provides the client for the V3 version of the AgentOps API.
"""

from agentops.client.api.base import BaseApiClient
from agentops.client.api.types import AuthTokenResponse
from agentops.exceptions import ApiServerException
from agentops.logging import logger
from termcolor import colored


class V3Client(BaseApiClient):
    """Client for the AgentOps V3 API"""

    def __init__(self, endpoint: str):
        """
        Initialize the V3 API client.

        Args:
            endpoint: The base URL for the API
        """
        # Set up with V3-specific auth endpoint
        super().__init__(endpoint)

    def fetch_auth_token(self, api_key: str) -> AuthTokenResponse:
        path = "/v3/auth/token"
        data = {"api_key": api_key}
        headers = self.prepare_headers()

        r = self.post(path, data, headers)

        if r.status_code != 200:
            error_msg = f"Authentication failed: {r.status_code}"
            try:
                error_data = r.json()
                if "error" in error_data:
                    error_msg = f"{error_data['error']}"
            except Exception:
                pass
            logger.error(f"{error_msg} - Perhaps an invalid API key?")
            raise ApiServerException(error_msg)

        try:
            jr = r.json()
            token = jr.get("token")
            if not token:
                raise ApiServerException("No token in authentication response")

            # Check project premium status
            if jr.get("project_prem_status") != "pro":
                logger.info(
                    colored(
                        "\x1b[34mYou're on the agentops free plan 🤔\x1b[0m",
                        "blue",
                    )
                )

            return jr
        except Exception as e:
            logger.error(f"Failed to process authentication response: {str(e)}")
            raise ApiServerException(f"Failed to process authentication response: {str(e)}")

    # Add V3-specific API methods here



import pytest
from unittest.mock import patch, MagicMock

# Tests for the new session management functionality
# These tests call the actual public API but mock the underlying implementation
# to avoid making real API calls or initializing the full telemetry pipeline


@pytest.fixture(scope="function")
def mock_tracing_core():
    """Mock the global tracer to avoid actual initialization"""
    # Patch both the main location and where it's imported in client
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.client.client.tracer", mock_tracer),
        patch("agentops.sdk.decorators.factory.tracer", mock_tracer),
        patch("agentops.legacy.tracer", mock_tracer),
    ):
        # Create a mock instance
        mock_tracer.initialized = True

        yield mock_tracer


@pytest.fixture(scope="function")
def mock_api_client():
    """Mock the API client to avoid actual API calls"""
    with patch("agentops.client.api.ApiClient") as mock_api:
        # Configure the v3.fetch_auth_token method to return a valid response
        mock_v3 = MagicMock()
        mock_v3.fetch_auth_token.return_value = {"token": "mock-jwt-token", "project_id": "mock-project-id"}
        mock_api.return_value.v3 = mock_v3

        yield mock_api


@pytest.fixture(scope="function")
def mock_trace_context():
    """Mock the TraceContext creation"""
    mock_span = MagicMock()
    mock_token = MagicMock()
    mock_trace_context_instance = MagicMock()
    mock_trace_context_instance.span = mock_span
    mock_trace_context_instance.token = mock_token
    mock_trace_context_instance.is_init_trace = False

    return mock_trace_context_instance


@pytest.fixture(scope="function")
def reset_client():
    """Reset the AgentOps client before each test"""
    import agentops

    # Create a fresh client instance for each test
    agentops._client = agentops.Client()
    # Reset all client state
    agentops._client._initialized = False
    agentops._client._init_trace_context = None
    agentops._client._legacy_session_for_init_trace = None
    yield
    # Clean up after test
    try:
        if hasattr(agentops._client, "_initialized"):
            agentops._client._initialized = False
        if hasattr(agentops._client, "_init_trace_context"):
            agentops._client._init_trace_context = None
        if hasattr(agentops._client, "_legacy_session_for_init_trace"):
            agentops._client._legacy_session_for_init_trace = None
    except:
        pass


def test_explicit_init_then_explicit_start_trace(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test explicitly initializing followed by explicitly starting a trace"""
    import agentops

    # Explicitly initialize with auto_start_session=False
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Verify that no auto-trace was started
    mock_tracing_core.start_trace.assert_not_called()

    # Mock the start_trace method to return our mock trace context
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Explicitly start a trace
    trace_context = agentops.start_trace(trace_name="test_trace", tags=["test"])

    # Verify the trace was created
    mock_tracing_core.start_trace.assert_called_once_with(trace_name="test_trace", tags=["test"])
    assert trace_context == mock_trace_context


def test_auto_start_session_true(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test initializing with auto_start_session=True"""
    import agentops
    from agentops.legacy import Session

    # Mock the start_trace method to return our mock trace context
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Initialize with auto_start_session=True
    result = agentops.init(api_key="test-api-key", auto_start_session=True)

    # Verify a trace was auto-started
    mock_tracing_core.start_trace.assert_called_once()
    # init() should return a Session object when auto-starting a session
    assert isinstance(result, Session)
    # Check that the session's trace_context has the expected properties
    assert result.trace_context is not None


def test_auto_start_session_default(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test initializing with default auto_start_session behavior"""
    import agentops
    from agentops.legacy import Session

    # Mock the start_trace method to return our mock trace context
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Initialize without explicitly setting auto_start_session (defaults to True)
    result = agentops.init(api_key="test-api-key")

    # Verify that the client was initialized
    assert agentops._client.initialized
    # Since auto_start_session defaults to True, init() should return a Session object
    assert isinstance(result, Session)
    # Check that the session's trace_context has the expected properties
    assert result.trace_context is not None


def test_start_trace_without_init():
    """Test starting a trace without initialization triggers auto-init"""
    import agentops

    # Reset client for test
    agentops._client = agentops.Client()

    # Mock global tracer to be uninitialized initially, then initialized after init
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.client.client.tracer", mock_tracer),
        patch("agentops.sdk.decorators.factory.tracer", mock_tracer),
        patch("agentops.legacy.tracer", mock_tracer),
    ):
        mock_tracer.initialized = False

        # Mock the init function to simulate successful initialization
        with patch("agentops.init") as mock_init:

            def side_effect():
                # After init is called, mark global tracer as initialized
                mock_tracer.initialized = True

            mock_init.side_effect = side_effect
            mock_tracer.start_trace.return_value = None

            # Try to start a trace without initialization
            result = agentops.start_trace(trace_name="test_trace")

            # Verify that init was called automatically
            mock_init.assert_called_once()
            # Should return None since start_trace returned None
            assert result is None


def test_end_trace(mock_tracing_core, mock_trace_context):
    """Test ending a trace"""
    import agentops

    # End the trace
    agentops.end_trace(mock_trace_context, end_state="Success")

    # Verify end_trace was called on global tracer
    mock_tracing_core.end_trace.assert_called_once_with(trace_context=mock_trace_context, end_state="Success")


def test_session_decorator_creates_trace(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test that the @session decorator creates a trace-level span"""
    import agentops
    from agentops.sdk.decorators import session

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Reset the call count to start fresh
    mock_tracing_core.reset_mock()
    # Mock the start_trace and end_trace methods
    mock_tracing_core.start_trace.return_value = mock_trace_context

    @session(name="test_session", tags=["decorator_test"])
    def test_function():
        return "test_result"

    # Execute the decorated function
    result = test_function()

    # Verify the function executed successfully
    assert result == "test_result"

    # Verify that start_trace and end_trace were called
    # Note: The decorator might call start_trace multiple times due to initialization
    assert mock_tracing_core.start_trace.call_count >= 1
    assert mock_tracing_core.end_trace.call_count >= 1


def test_session_decorator_with_exception(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test that the @session decorator handles exceptions properly"""
    import agentops
    from agentops.sdk.decorators import session

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Reset the call count to start fresh
    mock_tracing_core.reset_mock()
    # Mock the start_trace method
    mock_tracing_core.start_trace.return_value = mock_trace_context

    @session(name="failing_session")
    def failing_function():
        raise ValueError("Test exception")

    # Execute the decorated function and expect an exception
    with pytest.raises(ValueError, match="Test exception"):
        failing_function()

    # Verify that start_trace was called
    assert mock_tracing_core.start_trace.call_count >= 1
    # Verify that end_trace was called
    assert mock_tracing_core.end_trace.call_count >= 1


def test_legacy_start_session_compatibility(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test that legacy start_session still works and calls tracer.start_trace"""
    import agentops
    from agentops.legacy import Session

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Reset the call count to start fresh
    mock_tracing_core.reset_mock()
    # Mock the start_trace method
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Start a legacy session
    session = agentops.start_session(tags=["legacy_test"])

    # Verify the session was created
    assert isinstance(session, Session)
    # Check that the session's trace_context has the expected properties
    assert session.trace_context is not None

    # Verify that tracer.start_trace was called
    # Note: May be called multiple times due to initialization
    assert mock_tracing_core.start_trace.call_count >= 1


def test_legacy_end_session_compatibility(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test that legacy end_session still works and calls tracer.end_trace"""
    import agentops
    from agentops.legacy import Session

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Reset the call count to start fresh
    mock_tracing_core.reset_mock()

    # Create a legacy session object
    session = Session(mock_trace_context)

    # End the session
    agentops.end_session(session)

    # Verify that tracer.end_trace was called
    mock_tracing_core.end_trace.assert_called_once_with(mock_trace_context, end_state="Success")


def test_no_double_init(mock_tracing_core, mock_api_client, reset_client):
    """Test that calling init multiple times doesn't reinitialize"""
    import agentops

    # Initialize once
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Track the call count
    call_count = mock_api_client.call_count

    # Call init again with the same API key
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Verify that API client wasn't constructed again
    assert mock_api_client.call_count == call_count


def test_client_initialization_behavior(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test basic client initialization behavior"""
    import agentops

    # Mock the start_trace method
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Test that initialization works
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Verify that the client was initialized
    assert agentops._client.initialized

    # The API client might not be called if already mocked at a higher level
    # Just verify that the initialization completed successfully

    # Test that calling init again doesn't cause issues
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Should still be initialized
    assert agentops._client.initialized


def test_multiple_concurrent_traces(mock_tracing_core, mock_api_client, reset_client):
    """Test that multiple traces can be started concurrently"""
    import agentops

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Create mock trace contexts for different traces
    mock_trace_context_1 = MagicMock()
    mock_trace_context_2 = MagicMock()

    # Mock start_trace to return different contexts
    mock_tracing_core.start_trace.side_effect = [
        mock_trace_context_1,
        mock_trace_context_2,
    ]

    # Start multiple traces
    trace1 = agentops.start_trace(trace_name="trace1", tags=["test1"])
    trace2 = agentops.start_trace(trace_name="trace2", tags=["test2"])

    # Verify both traces were created
    assert trace1 == mock_trace_context_1
    assert trace2 == mock_trace_context_2

    # Verify start_trace was called twice
    assert mock_tracing_core.start_trace.call_count == 2


def test_trace_context_properties(mock_trace_context):
    """Test that TraceContext properties work correctly"""
    from agentops.legacy import Session

    # Create a legacy session with the mock trace context
    session = Session(mock_trace_context)

    # Test that properties are accessible
    assert session.span == mock_trace_context.span
    assert session.token == mock_trace_context.token
    assert session.trace_context == mock_trace_context


def test_session_decorator_async_function(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test that the @session decorator works with async functions"""
    import agentops
    import asyncio
    from agentops.sdk.decorators import session

    # Initialize AgentOps
    agentops.init(api_key="test-api-key", auto_start_session=False)

    # Reset the call count to start fresh
    mock_tracing_core.reset_mock()
    # Mock the start_trace method
    mock_tracing_core.start_trace.return_value = mock_trace_context

    @session(name="async_test_session")
    async def async_test_function():
        await asyncio.sleep(0.01)  # Simulate async work
        return "async_result"

    # Execute the decorated async function
    result = asyncio.run(async_test_function())

    # Verify the function executed successfully
    assert result == "async_result"

    # Verify that start_trace and end_trace were called
    assert mock_tracing_core.start_trace.call_count >= 1
    assert mock_tracing_core.end_trace.call_count >= 1


def test_trace_context_creation():
    """Test that TraceContext can be created with proper attributes"""
    from agentops.sdk.core import TraceContext

    mock_span = MagicMock()
    mock_token = MagicMock()

    # Test creating a TraceContext
    trace_context = TraceContext(span=mock_span, token=mock_token, is_init_trace=False)

    assert trace_context.span == mock_span
    assert trace_context.token == mock_token
    assert trace_context.is_init_trace == False


def test_session_management_integration():
    """Test the integration between new and legacy session management"""
    import agentops

    # Reset client for test
    agentops._client = agentops.Client()

    # Test that we can use both new and legacy APIs together
    with (
        patch("agentops.tracer") as mock_tracer,
        patch("agentops.client.client.tracer", mock_tracer),
        patch("agentops.sdk.decorators.factory.tracer", mock_tracer),
        patch("agentops.legacy.tracer", mock_tracer),
    ):
        mock_tracer.initialized = True

        # Mock API client
        with patch("agentops.client.api.ApiClient") as mock_api:
            mock_v3 = MagicMock()
            mock_v3.fetch_auth_token.return_value = {"token": "mock-jwt-token", "project_id": "mock-project-id"}
            mock_api.return_value.v3 = mock_v3

            # Initialize AgentOps
            agentops.init(api_key="test-api-key", auto_start_session=False)

            # Reset call counts after initialization
            mock_tracer.reset_mock()

            # Create mock trace context
            mock_trace_context = MagicMock()
            mock_tracer.start_trace.return_value = mock_trace_context

            # Test new API
            trace_context = agentops.start_trace(trace_name="new_api_trace")
            assert trace_context == mock_trace_context

            # Test legacy API
            session = agentops.start_session(tags=["legacy"])
            # Check that the session's trace_context has the expected properties
            assert session.trace_context is not None

            # Test ending both
            agentops.end_trace(trace_context)
            agentops.end_session(session)

            # Verify calls were made
            assert mock_tracer.start_trace.call_count >= 2
            assert mock_tracer.end_trace.call_count >= 2


# CrewAI Backwards Compatibility Tests
# These tests ensure CrewAI integration patterns continue to work


def test_session_auto_start(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """Test auto-start session functionality for CrewAI compatibility"""
    import agentops
    from agentops.legacy import Session

    # Configure mocks for session initialization
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Pass a dummy API key for the test
    session = agentops.init(api_key="test-api-key", auto_start_session=True)

    assert isinstance(session, Session)


def test_crewai_backwards_compatibility(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """
    CrewAI needs to access:

    agentops.track_agent
    agentops.track_tool
    agentops.start_session
    agentops.end_session
    agentops.ActionEvent
    agentops.ToolEvent
    """
    import agentops
    from agentops.legacy import Session

    # Configure mocks
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Test initialization with API key
    agentops.init(api_key="test-api-key")

    # Test session management functions
    session = agentops.start_session(tags=["test", "crewai"])
    assert isinstance(session, Session)

    # Test that passing a string to end_session doesn't raise an error
    agentops.end_session("Success")  # This pattern is used in CrewAI

    # Test track_agent function exists and doesn't raise errors
    try:
        # Mock an agent object similar to what CrewAI would provide
        class MockAgent:
            def __init__(self):
                self.role = "Test Agent"
                self.goal = "Testing"
                self.id = "test-agent-id"

        agent = MockAgent()
        agentops.track_agent(agent)
    except Exception as e:
        assert False, f"track_agent raised an exception: {e}"

    # Test track_tool function exists and doesn't raise errors
    try:
        # Mock a tool object similar to what CrewAI would provide
        class MockTool:
            def __init__(self):
                self.name = "Test Tool"
                self.description = "A test tool"

        tool = MockTool()
        agentops.track_tool(tool, "Test Agent")
    except Exception as e:
        assert False, f"track_tool raised an exception: {e}"

    # Test events that CrewAI might use
    tool_event = agentops.ToolEvent(name="test_tool")
    action_event = agentops.ActionEvent(action_type="test_action")

    # Verify that record function works with these events
    agentops.record(tool_event)
    agentops.record(action_event)


def test_crewai_kwargs_pattern(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """
    Test the CrewAI < 0.105.0 pattern where end_session is called with only kwargs.

    In versions < 0.105.0, CrewAI directly calls:
    agentops.end_session(
        end_state="Success",
        end_state_reason="Finished Execution",
        is_auto_end=True
    )
    """
    import agentops
    from agentops.legacy import Session

    # Configure mocks
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Initialize with test API key
    agentops.init(api_key="test-api-key")

    # Create a session
    session = agentops.start_session(tags=["test", "crewai-kwargs"])
    assert isinstance(session, Session)

    # Test the CrewAI < 0.105.0 pattern - calling end_session with only kwargs
    agentops.end_session(end_state="Success", end_state_reason="Finished Execution", is_auto_end=True)

    # After calling end_session, creating a new session should work correctly
    # (this implicitly tests that the internal state is reset properly)
    new_session = agentops.start_session(tags=["test", "post-end"])
    assert isinstance(new_session, Session)


def test_crewai_kwargs_pattern_no_session(mock_tracing_core, mock_api_client, reset_client):
    """
    Test the CrewAI < 0.105.0 pattern where end_session is called with only kwargs,
    but no session has been created.

    This should log a warning but not fail.
    """
    import agentops

    # Initialize with test API key
    agentops.init(api_key="test-api-key")

    # We don't need to explicitly clear the session state
    # Just make sure we start with a clean state by calling init

    # Test the CrewAI < 0.105.0 pattern - calling end_session with only kwargs
    # when no session exists. This should not raise an error.
    agentops.end_session(end_state="Success", end_state_reason="Finished Execution", is_auto_end=True)


def test_crewai_kwargs_force_flush(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """
    Test that when using the CrewAI < 0.105.0 pattern (end_session with kwargs),
    the spans are properly exported to the backend with force_flush.

    This is a more comprehensive test that ensures spans are actually sent
    to the backend when using the CrewAI integration pattern.
    """
    import agentops
    import time

    # Configure mocks
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Initialize AgentOps with API key
    agentops.init(api_key="test-api-key")

    # Create a session
    agentops.start_session(tags=["test", "crewai-integration"])

    # Simulate some work
    time.sleep(0.1)

    # End session with kwargs (CrewAI < 0.105.0 pattern)
    agentops.end_session(end_state="Success", end_state_reason="Test Finished", is_auto_end=True)

    # Explicitly ensure the core isn't already shut down for the test
    # Note: We're using mocks, so we check the mock's initialized state
    assert mock_tracing_core.initialized, "Mocked tracer should still be initialized"


def test_crewai_task_instrumentation(mock_tracing_core, mock_api_client, mock_trace_context, reset_client):
    """
    Test the CrewAI task instrumentation focusing on span attributes and tags.
    This test verifies that task spans are properly created with correct attributes
    and tags without requiring a session.
    """
    import agentops
    from opentelemetry.trace import SpanKind
    from agentops.semconv import SpanAttributes, AgentOpsSpanKindValues
    from opentelemetry import trace
    from agentops.semconv.core import CoreAttributes

    # Configure mocks
    mock_tracing_core.start_trace.return_value = mock_trace_context

    # Initialize AgentOps with API key and default tags
    agentops.init(
        api_key="test-api-key",
    )
    agentops.start_session(tags=["test", "crewai-integration"])
    # Get the tracer
    tracer = trace.get_tracer(__name__)

    # Create a mock task instance
    class MockTask:
        def __init__(self):
            self.description = "Test Task Description"
            self.agent = "Test Agent"
            self.tools = ["tool1", "tool2"]

    task = MockTask()

    # Start a span for the task
    with tracer.start_as_current_span(
        f"{task.description}.task",
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.AGENTOPS_SPAN_KIND: AgentOpsSpanKindValues.TASK.value,
            CoreAttributes.TAGS: ["crewai", "task-test"],
        },
    ) as span:
        # Verify span attributes
        assert span.attributes[SpanAttributes.AGENTOPS_SPAN_KIND] == AgentOpsSpanKindValues.TASK.value
        assert "crewai" in span.attributes[CoreAttributes.TAGS]
        assert "task-test" in span.attributes[CoreAttributes.TAGS]

        # Verify span name
        assert span.name == f"{task.description}.task"

        # Verify span kind
        assert span.kind == SpanKind.CLIENT

    agentops.end_session(end_state="Success", end_state_reason="Test Finished", is_auto_end=True)



from datetime import datetime, timezone


def get_ISO_time():
    """
    Get the current UTC time in ISO 8601 format with milliseconds precision in UTC timezone.

    Returns:
        str: The current UTC time as a string in ISO 8601 format.
    """
    return datetime.now(timezone.utc).isoformat()
